# Copyright (C) 2018-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

# flake8: noqa
# mypy: ignore-errors

from openvino.frontend.pytorch.py_pytorch_frontend import _FrontEndPytorchDecoder as Decoder
from openvino.frontend.pytorch.py_pytorch_frontend import _Type as DecoderType
from openvino.runtime import op, PartialShape, Type as OVType, OVAny, Shape
from openvino.frontend.pytorch.utils import maybe_convert_max_int, make_constant, fetch_attr, pt_to_ov_type_map, ov_to_c_type_map

import torch
import ctypes

class TorchFXPythonDecoder (Decoder):

    def __init__(self, pt_module, fx_gm, nodes=None, mark_node_callback=None, input_shapes=[], input_types=[]):
        Decoder.__init__(self)
        self.mark_node_callback = mark_node_callback
        # We store every decoder created by this decoder so that all them are not deleted until the first decoder is deleted
        self.m_decoders = []
        self.pt_module = pt_module
        self.fx_gm = fx_gm
        self.input_types = input_types
        self.input_shapes = input_shapes

        if issubclass(type(pt_module), torch.fx.graph_module.GraphModule):

            self._nodes = list(pt_module.graph.nodes)
            self._inputs = []
            self._outputs = []
            for i in range(len(self._nodes)):
                if self._nodes[i].op == 'placeholder':
                    self._inputs.append(i)
                elif self._nodes[i].op == 'output':
                    # Instead of putting output index, refer to its target
                    args = self._nodes[i].args
                    if isinstance(args[0], tuple):
                        args = args[0]
                    for output in args:
                        self._outputs.append(self._nodes.index(output))

        elif issubclass(type(pt_module), torch.fx.Node):


            self._nodes = nodes # passed from outer context

            # FIXME: Quadratic complexity nodes*nodes considering the outer loop over all nodes
            for i in range(len(self._nodes)):
                if self._nodes[i] == pt_module:
                    self._outputs = [i]

            # code constant or None input as a tuple to diffirentiate it from regular input
            # negative index is used to mark inputs initialized by inline constants, there are no such inputs in the graph
            self._inputs = [self._nodes.index(arg) if arg in self._nodes else (arg,) for arg in pt_module.args]

            # FIXME: Find a better way to pass nested tuples to OV frontend. This is a temprary solution to flatten arguments.
            new_inputs = []
            for i in range(len(pt_module.args)):
                expand_list = False
                if isinstance(pt_module.args[i], list):
                    for arg in pt_module.args[i]:
                        if arg in self._nodes:
                            expand_list = True
                            break;
                if expand_list:
                    for arg in pt_module.args[i]:
                        new_inputs.append(self._nodes.index(arg))
                else:
                    new_inputs.append(self._inputs[i])
            self._inputs = new_inputs



    def inputs(self):
        return [x if x is not None else 100000 for x in self._inputs]

    def input(self, index):  # TODO: remove
        return self.inputs()[index]  # TODO: find specialized method

    def get_input_debug_name(self, index):
        return "input"+str(index)

    def get_input_signature_name(self, index: int) -> str:
        return self.get_input_debug_name(index)

    def get_input_shape(self, index):
        if index < len(self.input_shapes):
            return PartialShape(self.input_shapes[index])
        input = self._raw_input(index)
        return self.get_shape_for_value(input)

    def get_input_type(self, index):
        if index < len(self.input_types):
            return OVAny(pt_to_ov_type_map[str(self.input_types[index])])
        input = self._raw_input(index)
        return self.get_type_for_value(input)

    def get_output_debug_name(self, index):
        return "output"+str(index)

    def get_output_shape(self, index):
        output = self._raw_output(index)
        return self.get_shape_for_value(output)

    def get_output_type(self, index):
        output = self._raw_output(index)
        return self.get_type_for_value(output)

    def _get_known_type_for_value(self, type):
        '''
            Returns known/unknown types wrapped as OVAny
        '''
        # Check for simple scalar types first
        # TODO: Don't use str, use native types
        if type is None:
            return OVAny(OVType.dynamic)
        if str(type) in pt_to_ov_type_map:
            return OVAny(pt_to_ov_type_map[str(type)])
        elif type.__class__ is torch.TensorType:
            # Tensor type, parse element type
            # TODO: replace string by native type
            # return OVAny(PartialShape([1,2,3]))
            return OVAny(DecoderType.Tensor(self._get_known_type_for_value(type.dtype())))
        elif type.__class__ is torch.ListType:
            element_type = type.getElementType()
            return OVAny(DecoderType.List(self._get_known_type_for_value(element_type)))
        else:
            # Not yet recognized
            return OVAny(OVType.dynamic)
            #pt_type_class = value.type().__class__
            #    if pt_type_class is torch.ListType:

    def get_shape_for_value(self, value):
        if value and ('tensor_meta' in value.meta.keys()):
            return PartialShape(value.meta['tensor_meta'].shape)
        return PartialShape([1])

    def get_type_for_value(self, value):
        if issubclass(type(value), torch.fx.Node):
            if ('tensor_meta' in value.meta.keys()):
                pt_type = value.meta['tensor_meta'].dtype
                if str(pt_type) in pt_to_ov_type_map:
                    ov_type = pt_to_ov_type_map[str(pt_type)]
                    return OVAny(ov_type)
            else:
                return OVAny(OVType.f32)
        elif isinstance(value, int):
            return OVAny(OVType.i32)
        elif isinstance(value, float):
            return OVAny(OVType.f32)
        elif isinstance(value, bool):
            return OVAny(OVType.boolean)
        else:
            return OVAny(OVType.f32)

    def get_subgraph_size(self):
        if issubclass(type(self.pt_module), torch.fx.Node):
            return 0
        return len(self.get_subgraphs()) if hasattr(self.pt_module, 'blocks') else 1

    def decoder_type_name(self) -> str:
        return "fx"

    def visit_subgraph(self, node_visitor):
        # make sure topological order is satisfied
        for node in self._nodes:
            if node.op == 'placeholder' or node.op == 'output':
                continue # skipping non-operational nodes
            decoder = TorchFXPythonDecoder(node, self.fx_gm, self._nodes, mark_node_callback=self.mark_node_callback)
            self.m_decoders.append(decoder)
            node_visitor(decoder)

    def get_subgraphs(self):
        if issubclass(type(self.pt_module), torch.fx.Node):
            return []
        return list(self.pt_module.blocks())

    def get_subgraph_decoder(self, index):
        decoder = TorchFXPythonDecoder(self.get_subgraphs()[index], self.fx_gm, mark_node_callback=self.mark_node_callback)
        self.m_decoders.append(decoder)
        return decoder

    def get_op_type(self):
        if self.pt_module.op == 'call_function':
            return str(self.pt_module.target)
        elif self.pt_module.op == 'get_attr':
            return 'get_attr'  # FIXME should be aligned with get_attr from TS implementation
        else:
            return 'UNKNOWN_TYPE_' + str(self.pt_module.op)

    def get_schema(self):
        return ''
        return self.pt_module.schema()

    def outputs(self):
        return self._outputs

    def _raw_outputs(self):
        return [self._nodes[x] for x in self._outputs]

    def _raw_output(self, index):
        return self._raw_outputs()[index]

    def _raw_inputs(self):
        return [self._nodes[x] if x is not None and x < len(self._nodes) else x for x in self._inputs]

    def _raw_input(self, index):
        return self._raw_inputs()[index]

    def num_of_outputs(self):
        return len(self.outputs())

    def output(self, index):
        return self.outputs()[index]

    def mark_node(self, node):
        if self.mark_node_callback is not None:
            self.mark_node_callback(self, node)
        return node

    def as_constant(self):

        if self.pt_module.op == 'get_attr':
            # Extract Constant from FX module field
            ret = fetch_attr(self.fx_gm, self.pt_module.target)
            ovshape = PartialShape(ret.size())
            ovtype = pt_to_ov_type_map[str(ret.type())]
            c_type = ctypes.POINTER(ov_to_c_type_map[ovtype])
            data_c_ptr = ctypes.cast(ret.data_ptr(), c_type)
            ov_const = op.Constant(ovtype, ovshape.get_shape(), data_c_ptr[:ret.nelement()])
            return ov_const.outputs()


        if not self.get_op_type() == 'prim::Constant':
            return None
        pt_value = self._raw_output(0)

        pt_type_class = pt_value.type().__class__
        if pt_type_class is torch.TensorType:
            return self.as_constant_tensor(pt_value)
        if pt_type_class is torch.ListType:
            return self.as_constant_list(pt_value)
        if str(pt_value.type()) in ['torch.int32', 'int']:
            return make_constant(OVType.i32, Shape([]), [pt_value.toIValue()]).outputs()
        if str(pt_value.type()) in ['torch.float', 'torch.FloatType', 'float']:
            return make_constant(OVType.f32, Shape([]), [pt_value.toIValue()]).outputs()
        if str(pt_value.type()) in ['torch.bool', 'bool']:
            return make_constant(OVType.boolean, Shape([]), [pt_value.toIValue()]).outputs()

        return None

    def as_string(self):
        if not self.get_op_type() == 'prim::Constant':
            return None
        pt_value = self._raw_output(0)

        if str(pt_value.type()) in ['torch.StringType', 'str']:
            return pt_value.toIValue()
        return None

    def as_constant_tensor(self, pt_value):
        ivalue = pt_value.toIValue()
        if pt_value.isCompleteTensor():
            try:
                ivalue = ivalue.to(memory_format=torch.contiguous_format).detach().cpu()
            except:
                print("[ WARNING ] Tensor couldn't detach")
            if str(pt_value.type().dtype()) in pt_to_py_type_map:
                # Constant interpretation doesn't respect new-full type of PT
                # It recognizes only tensors, and give lists as 1D tensors, and scalars as Tensor scalars
                # So only tensor-type constants are supported
                ovshape = PartialShape(pt_value.type().sizes())
                ovtype = pt_to_ov_type_map[str(pt_value.type().dtype())]

                # TODO: try-except here is a temporary WA for issues with data_ptr that we currently cannot predict; provide better solution
                try:
                    # this is only possible with adding a new ctor for Constant Python binding
                    # TODO Check strides and pass them somehow
                    values = ivalue.data_ptr()
                    ov_const = make_constant(ovtype, ovshape.get_shape(), values)
                except:
                    # old variant that makes a slow data copying
                    print(f"[ WARNING ] Constant wasn't able to convert from data_ptr.")
                    values = ivalue.flatten().tolist()
                    ov_const = make_constant(ovtype, ovshape.get_shape(), values)
                return ov_const.outputs()
        else:
            # Incomplete tensor can be scalar
            if isinstance(ivalue, float):
                return make_constant(OVType.f32, Shape([]), [ivalue]).outputs()
            if isinstance(ivalue, int):
                return make_constant(OVType.i32, Shape([]), [ivalue]).outputs()
            if isinstance(ivalue, bool):
                return make_constant(OVType.boolean, Shape([]), [ivalue]).outputs()

            # TODO: verify that it correctly reads incomplete consts
            if str(ivalue.type()) in pt_to_ov_type_map:
                try:
                    ovshape = PartialShape(ivalue.size())
                    ovtype = pt_to_ov_type_map[str(ivalue.type())]
                    ov_const = make_constant(ovtype, ovshape.get_shape(), ivalue.data_ptr())
                except:
                    # old variant that makes a slow data copying
                    print(f"[ WARNING ] Constant wasn't able to convert from data_ptr.")
                    nvalues = ivalue.numpy()
                    ovtype = np_to_ov_type_map[str(nvalues.dtype)]
                    ovshape = PartialShape(nvalues.shape)
                    ov_const = make_constant(ovtype, ovshape.get_shape(), nvalues.flatten().tolist())
                return ov_const.outputs()
        return None

    def as_constant_list(self, pt_value):
        # For now it is treat a list as a 1D tensor; it is required by converters to avoid need to massively rewrite them in that part where constant attributes are queried
        pt_element_type = str(pt_value.type().getElementType())
        ivalue = pt_value.toIValue()
        is_known_type = pt_element_type in pt_to_ov_type_map

        # WA to broken ov.Type
        # Detect integer list and process it with a dedicated method
        # TODO: Fix ov.Type and remove this WA
        # if pt_to_py_type_map[pt_element_type] == 'int':
        #    self.as_constant_list_of_ints(ovshape = PartialShape([len(ivalue)]), ivalue)
        # End of WA to broken ov.Type

        if is_known_type:
            ovtype = pt_to_ov_type_map[pt_element_type]
            ovshape = PartialShape([len(ivalue)])
            ov_const = make_constant(ovtype, ovshape.get_shape(), ivalue)
            return ov_const.outputs()

    def input_is_none(self, index):
        if index >= len(self.inputs()) or self._raw_input(index) is None:
            return True
        else:
            r_input = self._raw_input(index)
            return str(type(r_input)) in ['torch.NoneType', 'NoneType']

    def debug(self):
        self.pt_module.print()

    def inlined_inputs(self, index):
        result = []
        for i in range(len(self._inputs)):
            if isinstance(self._inputs[i], tuple):
                constant = None
                arg = self._inputs[i][0]
                if isinstance(arg, list):
                    if len(arg) > 0:
                        constant = make_constant(pt_to_ov_type_map[type(arg[0]).__name__], Shape([len(arg)]), arg)
                    else:
                        # TODO: which type should we use if list is empty? Need a signaling value here
                        constant = make_constant(int, Shape([0]), [])
                elif isinstance(arg, bool):
                    constant = make_constant(OVType.boolean, Shape([]), [arg])
                elif isinstance(arg, int):
                    arg = maybe_convert_max_int(arg)
                    constant = make_constant(OVType.i32, Shape([]), [arg])  # TODO: i32? why not i64?
                elif isinstance(arg, float):
                    constant = make_constant(OVType.f32, Shape([]), [arg])  # TODO: f32? why not f64?

                if constant is None:
                    if arg is None:
                        self._inputs[i] = None
                else:
                    assert len(constant.outputs()) == 1
                    result.append(constant.outputs()[0])
                    self._inputs[i] = index
                    index += 1
        return result

    def may_produce_alias(self, in_index: int, out_index: int) -> bool:
        if self.get_op_type() in ["aten::conv1d", "aten::conv2d", "aten::conv3d"]:
            # AliasDB::may_contain_alias sometimes return True for tensors produced by convnd, we have to workaround that
            return False
        try:
            return self.alias_db.may_contain_alias(self._raw_input(in_index), self._raw_output(out_index))
        except:
            # Sometimes pytorch fails to get result with IndexError exception while these indexes exist in node
            return False
